!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  
! 

  program bg
   use gmodel
   implicit none

   type :: part
    real:: posvel(0:7)
   end type part
!
   type(part),allocatable :: rdata(:),rdata1(:)

   real(kind=4):: rpar(10)
   integer :: ipar(10)

  integer:: npt,nps,nph,nth1,npg,nout,ngal,f_in,jam,nang,nmult,ncl
  integer:: nbd_tot,mods,modh,ncomp
  integer,allocatable ::ndim_gal(:),ipar_gal(:,:),rpar_gal(:,:)

  real :: virh,rah,rch,gamh,rh_max,Omh
  real :: vir,ra,rcs,gam,bet,rs_max,Oms
  real :: rcl,rmaxcl,phi,phiv
  real::  xvg(6,20)
  real :: totms,massp,epsbh

  real :: tnow,tdyn,ekin,epot,tote
  real,parameter :: gravc=1.0

  call data_in

  call init(nout)
  contains
!-------------------
  subroutine data_in
   integer,parameter:: lun=99
   integer :: k

   open(lun,file='newgalaxy.data')
   read(lun,*)
   read(lun,*)npt, nps,nrt
   read(lun,*)
   read(lun,*)mods,totms
   read(lun,*)
   read(lun,*)ra,rcs,gam
   close(lun)
   nout=0
   ipar=0
   rpar=0.0

!---------------------
   nph=64
   nth1=32
   nmult=nph*nth1

   ipar(1)=npt*nmult
   allocate(wp(nrt+1,5))
!
  end subroutine data_in

  subroutine init(id_new)
   integer,intent(in):: id_new
   integer::   nps,ng1,ng2
   integer::    i1,iout,ikv,k,nn
   real:: gam1,gam0,rot,ra2,r0,r1
   real:: rh,rhoh,mass_loc,ep_loc,fact
   real  :: x(3),v(3),Tx(3),Tv(6)

   integer,parameter:: lun=99

!
   iout=id_new
   tnow=0.0
   nbd_tot=ipar(1)
   nps=nbd_tot
   allocate(rdata(nbd_tot))
!------------------- now set initial data

    ng1=1
    ekin=0.0
    epot=0.0
   pi=acos(-1.d0)
   pi2=2.d0*pi
   pi4=2.d0*pi2

!!!!!!!!!!!!!!!!!!!!             ! star component
  if(nps/=0)then
    massp=totms/nps
    ng2=nps

    do i1=ng1,ng2
     rdata(i1)%posvel(0) = massp
     rdata(i1)%posvel(4) =i1
    end do                        !-----------------------------------------
!---------------------------
    call star_system(ng1,ng2,nps,rh,epot)

    epot=0.5*massp*epot
    if(Oms /= 0.0)then
     do i1=ng1,ng2
      x(1:3)=rdata(i1)%posvel(1:3)
      v(1:2)=rdata(i1)%posvel(5:6)
      r1=sqrt(x(1)*x(1)+x(2)*x(2))/rcs
      rot=Oms/(1.0+r1)**2
      
      v(2)=v(2)+rot*x(1) 
      v(1)=v(1)-rot*x(2) 
      rdata(i1)%posvel(5:6)=v(1:2)
     end do

    endif 
    call center(ng1,ng2,x,v,Tx,Tv)
    v=massp*v
    x=0.5*massp*x
    Tx=massp*Tx
    Tv=massp*Tv
    Tv(1:3)=0.5*Tv(1:3)
    ekin=sum(Tv(1:3))
    tote=ekin-epot
!------------------
  endif                          ! exit star component
!-----------------
    open(lun,file='init.dat')
     write(lun,*)'galaxy model'
     write(lun,'(a20,i8)')'number of particles= ',nbd_tot
     write(lun,'(a14,e11.4)')'tatal mass= ',totms
      select case(mods)
      case(1)
       write(lun,*)'Newtonian equilibrium gamma Model '
       write(lun,*)'     gam,      ra,      bet   '
       write(lun,'(4f10.3)')gam,ra,bet
      end select
!-------------------
       write(lun,*)'ekin  =',ekin,' epot=  ',epot, ' tote= ',tote
       write(lun,*)' half radius  rh=',rh
       write(lun,*)' dynam time=  ',tdyn
!       write(lun,*)'Lx= ',v(1),' Ly= ',v(2), ' Lz= ',v(3)
!       write(lun,*)'Txx= ',x(1),' Tyy= ',x(2), ' Tzz= ',x(3)
!       write(lun,*)'Txy= ',Tx(1),' Txz= ',Tx(2), ' Tyz= ',Tx(3)
!       write(lun,*)'Sxx= ',Tv(1),' Syy= ',Tv(2), ' Szz= ',Tv(3)
!       write(lun,*)'Sxy= ',Tv(4),' Sxz= ',Tv(5), ' Syz= ',Tv(6)
      close(lun)

     call out_part_data(id_new)

     deallocate(rdata)
!---------------------
  end subroutine init
  
  subroutine star_system(ng1,ng2,nps,rh,epot)

    integer,intent(in) :: ng1,ng2,nps
    real,intent(out) :: rh,epot
    real :: ra2,r0,r1,ep_loc,x(3),v(3)
    real   :: totmh,gam0,fact,phi,dteta,dphi,vs,vp
    integer :: i1,i,j,k,nn,ikv,nrg
    real,allocatable,dimension(:) :: cph,sph,cth,sth

    idum=-1

    select case(mods)
    case(1)
       write(*,*)'Newtonian equilibrium gamma Model '
       write(*,*)'     gam,      ra,      bet   '
       write(*,'(4f10.3)')gam,ra,bet
     ra2=ra*ra
     r0=rcs
     r1=rch
     totmh=0.0
     gamh=0.0
      call Hrnqst_par( &
                      totms,totmh,gam,gamh,ra2,r0,r1,tdyn,rh,epot,nrg)
      epot=0.0
      k=1
     do i1=ng1,ng2
      call OM_dist(k,ra2,r0,x,v,ep_loc)
      epot=epot+ep_loc
       rdata(i1)%posvel(1:3) =x(1:3)
       rdata(i1)%posvel(5:7) =v(1:3)
     k=-k
     end do
    end select
     write(*,'(a20,i8)')'number of particles= ',nbd_tot
     write(*,'(a14,e11.4)')'total mass= ',totms
  end subroutine star_system

  subroutine center(n1,n2,xx,vv,Tx,Tv)
   integer  :: n1,n2,k
   real :: xx(3),vv(3),Tx(3),Tv(6),norm

   norm=1.0/float(n2-n1+1)
      do k=1,3
       xx(k)=sum(rdata(n1:n2)%posvel(k))      
       vv(k)=sum(rdata(n1:n2)%posvel(k+4))      
      end do
      xx=xx*norm
      vv=vv*norm
    do k=n1,n2
     rdata(k)%posvel(1:3)=rdata(k)%posvel(1:3)-xx(1:3)
     rdata(k)%posvel(5:7)=rdata(k)%posvel(5:7)-vv(1:3)
    end do
      do k=1,3
       xx(k)=sum(rdata(n1:n2)%posvel(k))      
       vv(k)=sum(rdata(n1:n2)%posvel(k+4))      
      end do
      xx=xx*norm
      vv=vv*norm
    
    vv=0.0
    Tv=0.0
    do k=n1,n2
     xx(1:3)=rdata(k)%posvel(5:7)
     Tv(1:3)=Tv(1:3)+xx(1:3)*xx(1:3)
     Tv(4)=Tv(4)+xx(1)*xx(2)
     Tv(5)=Tv(5)+xx(1)*xx(3)
     Tv(6)=Tv(6)+xx(3)*xx(2)
    end do
    xx=0.0
    Tx=0.0
    do k=n1,n2
     xx(1:3)=xx(1:3)+rdata(k)%posvel(1:3)*rdata(k)%posvel(1:3)
     Tx(1)=Tx(1)+rdata(k)%posvel(1)*rdata(k)%posvel(2)
     Tx(2)=Tx(2)+rdata(k)%posvel(1)*rdata(k)%posvel(3)
     Tx(3)=Tx(3)+rdata(k)%posvel(2)*rdata(k)%posvel(3)
    end do
    vv=0.0
    do k=n1,n2
     vv(1)=vv(1)+rdata(k)%posvel(2)*rdata(k)%posvel(7)
     vv(1)=vv(1)-rdata(k)%posvel(3)*rdata(k)%posvel(6)
     vv(2)=vv(2)+rdata(k)%posvel(3)*rdata(k)%posvel(5)
     vv(2)=vv(2)-rdata(k)%posvel(1)*rdata(k)%posvel(7)
     vv(3)=vv(3)+rdata(k)%posvel(1)*rdata(k)%posvel(6)
     vv(3)=vv(3)-rdata(k)%posvel(2)*rdata(k)%posvel(5)
    end do
  end subroutine center

  subroutine out_part_data(iout)

  integer,intent(in):: iout
  integer:: i,nrec
  character(6) :: fname='     '
  real(kind=4) :: arr(1:6)

! Output written on file 'out0n.dat'
   rpar=0.0
   ipar=0
   rpar(1)=totms
   rpar(2)=tnow
   rpar(3)=tdyn
   rpar(4)=vir
   ipar(1)=nbd_tot
   ipar(2)=mods

!--------------
  if (iout<100) write (fname,'(a4,i2)') 'mout' ,iout
  if (iout< 10) write (fname,'(a5,i1)') 'mout0',iout
  print*,'Main parameters reported in file init.dat'
!----------------- allocate buffers for pe0

  open(10,file=fname//'.bin',form='unformatted')
   write(10)ipar(1:5)
   write(10)rpar(1:5)
   do i=1,nbd_tot
    arr(1:3)=rdata(i)%posvel(1:3)
    arr(4:6)=rdata(i)%posvel(5:7)
    write(10)arr(1:6)
   enddo
   close(10)
!--------------
  print*,'Output written on file: '//fname//'.bin'
  end subroutine out_part_data


!-------------
 end program bg
