!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  
! 

module gmodel
 implicit none

 integer :: nrt
 real(kind=8) :: pi,pi2,pi4

 real,allocatable :: den(:),pot(:),rgr(:),wp(:,:)
 real,allocatable :: trad(:), th(:),d0(:),d1(:),dth(:),pth(:)
 real             :: alpha,gamma,anr,q0,bh,om,rmax,epdata(8)
 real         :: aph1,aph2,aphinv,beta,anr2,y0,dth_inv
 integer :: nr,nlim,nth,idum
 real :: a,b,c,b2,aa2,vexp

contains
  subroutine Hrnqst_par(am1,am2,gam0,gam1, &
                         rsq,r0,r1,dyn,half,epot,ngr)

   real,intent(in):: am1,am2,gam0,gam1,rsq,r0,r1
   real,intent(out) :: dyn,half,epot
   integer,intent(out):: ngr
   integer  :: k,tpt
   real     :: gg,gg1,g3,u,xr,pot1,norm
   real     :: a2,s1,sa,sa2,rg,gfact,umax,dr

    wp=0.0

    ngr=nrt   
    umax=0.99
    dr=nrt/umax

    gg=2.0-gam0
    g3=3.0-gam0
    a2=1./g3
    rg=(0.5)**a2
    half=r0*rg/(1.0-rg)
    dyn=pi*half*sqrt(0.5*half/(am1+am2))
!
    do k=1,nrt
     u=float(k)/dr
     xr=u**a2
     xr=xr/(1.-xr)
     xr=r0*xr
     sa=xr/rsq
     sa2=1.+xr*sa
     s1=r0+xr
     rg=g3*(s1/xr)**gam0
     wp(k,1)=xr
     wp(k,4)=am1*rg/(s1**4)
     if(gam0 < 2.)then
      wp(k,3)=(1.-(xr/s1)**gg)/gg
     else
      wp(k,3)=alog(s1/xr)
     endif
     wp(k,3)=am1*wp(k,3)/r0
     wp(k,2)=sa2*wp(k,4)/(pi4)
    end do
    if(am2 /=0.0)then
     gg1=2.0-gam1
     do k=1,nrt
      xr=wp(k,1)
      s1=r1+xr
      if(gam1 < 2.)then
       pot1=(1.-(xr/s1)**gg1)/gg1
      else
       pot1=alog(s1/xr)
      endif
      wp(k,3)=wp(k,3)+am2*pot1/r1
     end do
    endif
!------------------
    epot=0.0
    do k=1,nrt
     u=float(k)/dr
     xr=u**a2
     gfact=a2/(u*(1.0-xr))
     xr=xr/(1.-xr)
     epot=epot+gfact*wp(k,3)*wp(k,4)*xr*xr*xr
    end do
    epot=0.5*epot/dr
!-----------------
     wp(:,4)=0.0
     call fqs(nrt)   
                !exits with fq(),wp(:,2) destroyed wp(:,1), wp(:3) preserved

     norm=1./(sqrt(2.)*pi*pi)
     wp(1:nrt,5)=wp(1:nrt,5)*norm
!     write(10,*)'rad'
!     write(10,'(5e11.4)')wp(1:nrt,1)
!     write(10,*)'pot'
!     write(10,'(5e11.4)')wp(1:nrt,3)
!     write(10,*)'fq'
!     write(10,'(5e11.4)')fq(1:nrt)

 end subroutine Hrnqst_par
!-------------------------------------------------------
!
  subroutine OM_dist(is,rsq,r0,x,v,eph)

   integer,intent(in) :: is
   real,intent(in) :: rsq,r0
   real,intent(out) :: x(3),v(3),eph

   real  :: u,du,r2
   integer::       k,j,ii,ncycle
   integer,parameter  :: ncycl_max=800
   real:: ph,cph,sph,cth,sth,s
   real :: vr,vth,vs,vph
   real:: sa,afact,psit,ut2,ur2,qv,ff,pp
     
     call ran_number(idum,u)
     u=(nrt-1)*u
     k=nint(u)
!     du=u-float(k)
     k=k+1
     s=wp(k,1)
     psit=wp(k,3)
     eph=psit
     r2=1.+s*s/rsq
      sa=sqrt(r2)
      afact=sqrt(2.*psit)
      ncycle=0
      do
       do
        call ran_number(idum,v(1))
        call ran_number(idum,v(2))
        call ran_number(idum,v(3))
        v=2.0*v-1.0
        vr=v(1)
	vth=v(2)
	vph=v(3)
	ur2=vr*vr
        ut2=vth*vth+vph*vph
        u=ur2+ut2
        if(u >= 1.0)cycle
        exit
       end do
!
        call ran_number(idum,pp)
        pp=wp(k,5)*pp

        qv=psit*(1.0-u)
        j=0
        do ii=k,nrt
         if(wp(ii,3)>=qv)j=ii
        end do
        if(j==0)then
       !  print *,k,qv,wp(k,3)
         exit
        endif
        ncycle=ncycle+1
        if(ncycle > ncycl_max)exit
!        qv=(qv-wp(j,3))/(wp(j+1,3)-wp(j,3)) 
!        ff=(1.0-qv)*wp(j,5)+qv*wp(j+1,5)
         ff=wp(j,5)
         if(pp > ff)cycle
       exit
      end do
      if(ncycle >ncycl_max)then
     !  print *,' k,   j,   qv,   wp(j,5),  wp(k,5)'
     !  print *,k,j,qv,wp(j,5),wp(k,5)
      endif

       vth=vth/sa
       vph=vph/sa
       vr=afact*vr
       vth=afact*vth
       vph=afact*vph
                                ! recover cartesian coordinates
       call ran_number(idum,cth)
       call ran_number(idum,ph)

       cth=is*cth
       sth=sqrt(1.-cth*cth)
       ph=pi2*ph
       cph=cos(ph)
       sph=sin(ph)
         x(1)=s*sth*cph
         x(2)=s*sth*sph
         x(3)=s*cth

         vs=vr*sth+vth*cth
         v(3)=vr*cth-vth*sth
         v(1)=vs*cph-vph*sph
         v(2)=vs*sph+vph*cph

  end subroutine OM_dist
!-------------------
  subroutine fqs(np)
   integer,intent(in):: np
   integer:: i,j,k
   real:: drpsi,dqs,dqi

    do j=1,np
     drpsi=(wp(j+1,2)-wp(j,2))/(wp(j+1,3)-wp(j,3))
     wp(j,4)=drpsi
    end do
    do i=1,np
     wp(i,2)=.0
    dqi=.0
    do j=i,np
     dqs=sqrt(wp(i,3)-wp(j+1,3))
     wp(i,2)=wp(i,2)+wp(j,4)*(dqs-dqi)
     dqi=dqs
    end do
    end do
    do i=1,np-1
     wp(i,5)=(wp(i+1,2)-wp(i,2))/(wp(i+1,3)-wp(i,3))
    end do
    i=1 
     do
      if(wp(i+1,5) > wp(i,5))then
        k=i
        exit
      endif
       if(i==np-2)exit
       i=i+1
     end do 
!     write(*,*)k,wp(k-1:k+1,5)
     do i=k,np-1
!      wp(i+1,5)=3.0*(wp(i,5)-wp(i-1,5))+wp(i-2,5)
      wp(i+1,5)=2.0*wp(i,5)-wp(i-1,5)
     end do 
  end subroutine fqs
!-------------------



  subroutine ran_number(idum,ran2)
   integer,intent(inout) :: idum
    real,intent(out)  :: ran2 

        integer,parameter :: im1=2147483563,im2=2147483399,ia1=40014, &
                              ia2=40692,iq1=53668,iq2=52774,ir1=12211,&
                              ir2=3791,ntab=32,imm1=im1-1,ndiv=1+imm1/ntab 
        real,parameter   ::  am=1.0/im1,eps=1.2e-07,rnmx=1.0-eps 
        integer          :: j,k
        integer,save   :: iv(32),iy,idum2
        data idum2/123456789/,iv/32*0/,iy/0/ 

        if( idum <= 0)then
         idum=max(-idum,1)
         idum2=idum
         do j=ntab+8,1,-1
          k=idum/iq1
          idum=ia1*(idum-k*iq1)-k*ir1
          if(idum < 0)idum=idum+im1
          if(j <=ntab)iv(j)=idum
         end do
         iy=iv(1)
        endif
        k=idum/iq1
        idum=ia1*(idum-k*iq1)-k*ir1
        if(idum <0)idum=idum+im1
        k=idum2/iq2
        idum2=ia2*(idum2-k*iq2)-k*ir2
        if(idum2 <0)idum2=idum2+im2
        j=1+iy/ndiv
        iy=iv(j)-idum2
        iv(j)=idum
        if(iy <1)iy=iy+imm1
        ran2=min(am*iy,rnmx)
     end subroutine ran_number

 end module gmodel
