!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  


 module fft_lib

 implicit none
 contains
! ***************************************************************************

   subroutine ftw_alloc(n3)
    integer,intent(in) :: n3
   end subroutine ftw_alloc

  subroutine ft1_ph(w,nx,ny,nz,is)
   integer,intent(in) :: nx,ny,nz,is

   real,intent(inout) :: w(0:nx,0:ny-1,0:nz-1)
   real :: sc
   integer :: ix,iy,iz
   real :: ur(1:2*nz)

    select case(is)
    case(-1)
     sc=1./float(nz)
     do iy=0,ny-1
      do ix=0,nx
       ur(1:2*nz-1:2)=w(ix,iy,0:nz-1)
       ur(2:2*nz:2)=0.
       call four1(ur,nz,is)
       w(ix,iy,0:nz-1)=sc*ur(1:nz)
      end do
     end do
    case(1)   
!    sc=float(nz)
    do iy=0,ny-1
     do ix=0,nx
        ur(1:nz)=w(ix,iy,0:nz-1)
        ur(nz+1:nz+2)=0.
        do iz=1,nz-3,2
           ur(2*nz-iz)=ur(iz+2)
           ur(2*nz+1-iz)=-ur(iz+3)
        enddo
        call four1(ur,nz,is)
        w(ix,iy,0:nz-1)=ur(1:2*nz-1:2)
     end do
    end do
   end select
  end subroutine ft1_ph


  !   subroutine four1 of Numerical Recipes (Press et al.)
      SUBROUTINE four1(data,nn,isign)
      INTEGER isign,nn
      REAL data(2*nn)
      INTEGER i,istep,j,m,mmax,n
      REAL tempi,tempr
      DOUBLE PRECISION theta,wi,wpi,wpr,wr,wtemp
      n=2*nn
      j=1
      do 11 i=1,n,2
        if(j.gt.i)then
          tempr=data(j)
          tempi=data(j+1)
          data(j)=data(i)
          data(j+1)=data(i+1)
          data(i)=tempr
          data(i+1)=tempi
        endif
        m=nn
1       if ((m.ge.2).and.(j.gt.m)) then
          j=j-m
          m=m/2
        goto 1
        endif
        j=j+m
11    continue
      mmax=2


2     if (n.gt.mmax) then
        istep=2*mmax
        theta=6.28318530717959d0/(isign*mmax)
        wpr=-2.d0*sin(0.5d0*theta)**2
        wpi=sin(theta)
        wr=1.d0
        wi=0.d0
        do 13 m=1,mmax,2
          do 12 i=m,n,istep
            j=i+mmax
            tempr=sngl(wr)*data(j)-sngl(wi)*data(j+1)
            tempi=sngl(wr)*data(j+1)+sngl(wi)*data(j)
            data(j)=data(i)-tempr
            data(j+1)=data(i+1)-tempi
            data(i)=data(i)+tempr
            data(i+1)=data(i+1)+tempi
12        continue
          wtemp=wr
          wr=wr*wpr-wi*wpi+wr
          wi=wi*wpr+wtemp*wpi+wi
13      continue
        mmax=istep
      goto 2
      endif
      return
      END SUBROUTINE four1

!-------------------------------------------

 end module fft_lib 
