!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  
! 

 module struct
 type::  particle
  real(kind=4) :: posvel(6)
 end type particle  
end module struct

module parallel
  use struct
  include 'mpif.h'

  integer :: partype

  integer :: mype,npe,pe_min,pe_max,pe_h,npe_h,kd,   &
             comm,status(mpi_status_size),error,mpi_sd

  integer,allocatable  :: pe_next(:),pe_prev(:)
  real,allocatable :: sbuff(:,:,:),lbdth(:,:,:),rbdth(:,:,:)

  integer,allocatable,dimension(:,:):: pe_ind

  integer,allocatable,dimension(:):: nbd

  integer :: nbd_max,nbd_min

  logical :: pe0,prl

 contains

 subroutine start_parallel
    integer:: k,ipe,ik2,k_h,k_hh


    call mpi_init(error)

    call mpi_comm_rank(mpi_comm_world,mype,error)

    call mpi_comm_size(mpi_comm_world,npe,error)

    comm=mpi_comm_world

    pe0=mype==0
    prl=npe>1

    pe_min=0
    pe_max=npe-1
    npe_h=npe/2
    pe_h=npe_h-1

    k=kind(1.0)
  
    if(k==4)mpi_sd=mpi_real
    if(k==8)mpi_sd=mpi_double_precision

    if(prl)then

    allocate(pe_next(npe-1),pe_prev(npe-1))

     do ipe=1,npe-1
      pe_next(ipe)=mod(ipe+mype,npe)
      if(mype >=ipe)then
       pe_prev(ipe)=mype-ipe
      else
       pe_prev(ipe)=mype-ipe+npe
      endif 
  
     end do

     call mpi_type_contiguous(6,mpi_real,partype,error)
     call mpi_type_commit(partype,error)

     kd=1
      do 
       ipe=2**kd
       if(ipe==npe)exit
       kd=kd+1
      end do
!
     allocate(pe_ind(0:pe_max,1:kd))
     
     pe_ind=1
     k_h=npe
      do ipe=1,kd
       k_hh=k_h/2
       do k=0,k_hh-1
        pe_ind(k:pe_max-1:k_h,ipe)=-1
       enddo
       k_h=k_h/2
      enddo
!
   endif

   allocate(nbd(0:pe_max))

  end subroutine start_parallel
!
  subroutine end_parallel

   call mpi_finalize(error)

  end subroutine end_parallel

  subroutine exchange_grdata(sr,dat0,n1,n2,n3,ipe,tag)
   integer,intent(in) :: n1,n2,n3,ipe,tag
   real,intent(inout) :: dat0(:,:,:)
   logical,intent(in) :: sr
   integer :: lenw

   lenw=n1*n2*n3
    if(sr)then

     call mpi_send(dat0(1,1,1),lenw,mpi_sd,ipe,tag, &
                        comm,error)

    else

     call mpi_recv(dat0(1,1,1),lenw,mpi_sd,ipe,tag, &
                        comm,status,error)
    endif

  end subroutine exchange_grdata

  subroutine exchange_thbd_data(dat0,n1,n2,n3,npt)
   integer,intent(in) :: n1,n2,n3,npt
   real,intent(in) :: dat0(0:n1,0:n2-1,0:n3-1)
   integer :: k,j,j1,lenw,ip,tag,spe,rpe

   lenw=npt*(n1+1)*n3
   ip=1
    spe=pe_next(ip)
    rpe=pe_prev(ip)
    tag=10+ip
    do k=0,n3-1
     do j=1,npt
      sbuff(1:n1+1,k+1,j)=dat0(0:n1,n2-j,k)
     end do
    end do
    
     call mpi_sendrecv(sbuff(1,1,1),lenw,mpi_sd,spe,tag, &
                       lbdth(1,1,1),lenw,mpi_sd,rpe,tag, &
                        comm,status,error)
    rpe=pe_next(ip)
    spe=pe_prev(ip)
    tag=20+ip
    do k=0,n3-1
     do j=0,npt-1
      sbuff(1:n1+1,k+1,j+1)=dat0(0:n1,j,k)
     end do
    end do
     call mpi_sendrecv(sbuff(1,1,1),lenw,mpi_sd,spe,tag, &
                       rbdth(1,1,1),lenw,mpi_sd,rpe,tag, &
                        comm,status,error)

  end subroutine exchange_thbd_data

  subroutine fill_grdata(dat0,aux1,aux2,n1,n2t,n2,n3)
   integer,intent(in) :: n1,n2t,n2,n3
   real,intent(in) :: aux1(n1+1,n2,n3)
   real,intent(inout) :: dat0(0:n1,0:n2t-1,0:n3-1)
   real :: aux2(n1+1,n2,n3)
   integer :: k,j1,j,lenw,ip,tag,spe,rpe

   lenw=(n1+1)*n2*n3

   do ip=1,npe-1
    spe=pe_next(ip)
    rpe=pe_prev(ip)
    tag=ip
     call mpi_sendrecv(aux1(1,1,1),lenw,mpi_sd,spe,tag, &
                       aux2(1,1,1),lenw,mpi_sd,rpe,tag, &
                        comm,status,error)
    do k=0,n3-1
     do j=0,n2-1
      j1=j+n2*rpe
      dat0(0:n1,j1,k)=aux2(1:n1+1,j+1,k+1)
     end do
    end do
   end do
  end subroutine fill_grdata

  subroutine intvec_dist(nc)

   integer :: nc(0:pe_max)
   integer :: ipe

   if(.not.pe0)then

    call mpi_send(nc(mype),1,mpi_integer,pe_min,20,comm,error)
   else
    do ipe=1,pe_max
     call mpi_recv(nc(ipe),1,mpi_integer,ipe,20,comm,status,error)
    end do
   endif

   call mpi_bcast(nc,npe,mpi_integer,pe_min,comm,error)

  end subroutine intvec_dist

  subroutine exchange_pdata(sr,dat0,lenw,ipe,tag)
   integer,intent(in) :: lenw,ipe,tag
   logical,intent(in) :: sr

   type(particle),intent(inout) :: dat0(lenw)

    if(sr)then

     call mpi_send(dat0(1),lenw,partype,ipe,tag, &
                        comm,error)
    else

     call mpi_recv(dat0(1),lenw,partype,ipe,tag, &
                        comm,status,error)
    endif

  end subroutine exchange_pdata

  subroutine exchange_rdata(sr,dat0,lenw,ipe,tag)
   integer,intent(in) :: lenw,ipe,tag
   logical,intent(in) :: sr

   real,intent(inout) :: dat0(lenw)

    if(sr)then

     call mpi_send(dat0(1),lenw,mpi_sd,ipe,tag, &
                        comm,error)

    else

     call mpi_recv(dat0(1),lenw,mpi_sd,ipe,tag, &
                        comm,status,error)
    endif

  end subroutine exchange_rdata

  subroutine exchange_idata(sr,dat0,lenw,ipe,tag)
   integer,intent(in) :: lenw,ipe,tag
   logical,intent(in) :: sr

   integer,intent(inout) :: dat0(lenw)

    if(sr)then

     call mpi_send(dat0(1),lenw,mpi_integer,ipe,tag, &
                        comm,error)

    else

     call mpi_recv(dat0(1),lenw,mpi_integer,ipe,tag, &
                        comm,status,error)
    endif

  end subroutine exchange_idata
! ***************************************************************************
  subroutine allreduce_real_sdata(dt_prl,dt,ib)

   integer,intent(in) :: ib
   real :: dt_prl
   real :: dt
   integer,parameter :: nt=1

   if(prl)then
    select case(ib)
    case(0)             !sum

    call mpi_reduce(dt_prl,dt,nt,mpi_sd,mpi_sum,pe_min,comm,error)
    case(1)             !min

    call mpi_reduce(dt_prl,dt,nt,mpi_sd,mpi_min,pe_min,comm,error)
    case(2)             !max

    call mpi_reduce(dt_prl,dt,nt,mpi_sd,mpi_max,pe_min,comm,error)
    end select
    call mpi_bcast(dt,nt,mpi_sd,pe_min,comm,error)
   else
    dt=dt_prl
   endif

  end subroutine allreduce_real_sdata 
  subroutine allreduce_real_data(nt,dt_prl,dt,ib)

   integer,intent(in) :: ib,nt
   real :: dt_prl(nt)
   real :: dt(nt)

   if(prl)then
    select case(ib)
    case(0)             !sum

    call mpi_reduce(dt_prl,dt,nt,mpi_sd,mpi_sum,pe_min,comm,error)
    case(1)             !min

    call mpi_reduce(dt_prl,dt,nt,mpi_sd,mpi_min,pe_min,comm,error)
    case(2)             !max

    call mpi_reduce(dt_prl,dt,nt,mpi_sd,mpi_max,pe_min,comm,error)
    end select
    call mpi_bcast(dt,nt,mpi_sd,pe_min,comm,error)
   else
    dt=dt_prl
   endif

  end subroutine allreduce_real_data 

  subroutine bcast_real_data(nt,dt)

   integer,intent(in) :: nt
   real :: dt(nt)

    call mpi_bcast(dt,nt,mpi_sd,pe_min,comm,error)

  end subroutine bcast_real_data

  subroutine allreduce_int_sdata(dt_prl,dt,mp_mode)

   integer,intent(in) :: mp_mode
   integer :: dt_prl
   integer :: dt
   integer,parameter :: nt=1

   if(prl)then
    select case(mp_mode)
     case(0)
     call mpi_reduce(dt_prl,dt,nt,mpi_integer,mpi_sum,pe_min,comm,error)
     case(1)
     call mpi_reduce(dt_prl,dt,nt,mpi_integer,mpi_min,pe_min,comm,error)
     case(2)
     call mpi_reduce(dt_prl,dt,nt,mpi_integer,mpi_max,pe_min,comm,error)
    end select
    call mpi_bcast(dt_prll,nt,mpi_integer,pe_min,comm,error)
   else
    dt=dt_prl
   endif
  end subroutine allreduce_int_sdata

 subroutine bcast_int_data(nt,dt)

  integer,intent(in) :: nt
  integer :: dt(nt)

  call mpi_bcast(dt,nt,mpi_integer,pe_min,comm,error)

 end subroutine bcast_int_data
!=======================
 subroutine allreduce_int_data(nt,dt_prl,dt,mp_mode)

   integer,intent(in) :: mp_mode,nt
   integer :: dt_prl(nt)
   integer :: dt(nt)

   if(prl)then
    select case(mp_mode)
     case(0)
     call mpi_reduce(dt_prl,dt,nt,mpi_integer,mpi_sum,pe_min,comm,error)
     case(1)
     call mpi_reduce(dt_prl,dt,nt,mpi_integer,mpi_min,pe_min,comm,error)
     case(2)
     call mpi_reduce(dt_prl,dt,nt,mpi_integer,mpi_max,pe_min,comm,error)
    end select
    call mpi_bcast(dt_prll,nt,mpi_integer,pe_min,comm,error)
   else
    dt=dt_prl
   endif

  end subroutine allreduce_int_data

  subroutine sumreduce_grdata(ib,dat0,dat1,n1,n2,n3)
   logical,intent(in) :: ib
   integer,intent(in) :: n1,n2,n3
   real,intent(in) :: dat0(n1,n2,n3) 
   real,intent(out) :: dat1(n1,n2,n3) 
   integer :: lenw

   lenw=n1*n2*n3

   call mpi_reduce(dat0(1,1,1),dat1(1,1,1),&
                   lenw,mpi_sd,mpi_sum,pe_min,comm,error)
   if(ib)call mpi_bcast(dat1(1,1,1),lenw,mpi_sd,pe_min,comm,error)

  end subroutine sumreduce_grdata
!----------------------------------
 end module parallel
