!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  
! 

 module particles
  use griddata
  implicit none

  real, allocatable :: xvp(:,:) 
  real, allocatable :: p_acc(:,:) 
  real :: sc,rx,ry,rz,pig
  integer :: rm,ord,nb
  real,parameter :: half=0.5,thr=0.75

  contains
!=============== Stores p_acc(:,0)=r_particle, xvp(:,0)=xi_particle
  subroutine inv_rad_map(np)
   integer,intent(in) :: np
   integer :: n
   real :: r2,r,ax

    select case(rm)
    case(1)             !r=tan(x)
    do n=1,np 
     r2=dot_product(xvp(n,1:3),xvp(n,1:3))
     r=sqrt(r2)
     xvp(n,0)=atan(r/sc)
     p_acc(n,0)=r
    end do

    case(2)             !r=tan^2(x)
    do n=1,np 
     r2=dot_product(xvp(n,1:3),xvp(n,1:3))
     r=sqrt(r2)
     ax=r/sc
     xvp(n,0)=atan(sqrt(ax))
     p_acc(n,0)=r
    end do
    end select
  end subroutine inv_rad_map
!=============================
! Assigns grid-based acceleration to particles positions by interpolation
!============================================= 
  subroutine part_acc(n1,n2,n3)
   integer,intent(in) :: n1,n2,n3

   integer :: n,i,i1,i2,j,j1,j2,k,k1,k2

    real :: ax,ax2,ay,az,r2c,r,phi,dvol
    real :: ax0(0:2),ay0(0:2),az0(0:2)
    real :: cn,sn,cnp,snp,fx,fz

!--------------------------------------
!   WARNING : (Gr=Dpot/Dr  Gth=Dpot/Dtheta  Gph=Dpot/Dphi
!              geometrical factors r,sin(th) are not interpolated
! The Cartesian components on a particle are then given by
! Gz=Gr*cos(th)-Gth*sin(th)/r
! GR=Gr*cos(th)+Gth*sin(th)/r
! Gx=GR*cos(ph)-Gph*sin(ph)/(r*sin(th)
! Gy=f_R*sin(ph)+Gph*cos(ph)/(r*sin(th))
!
!-------------------
    call inv_rad_map(nb)
!-------------------
   select case(ord)
!-----------------------
    case(1)               ! linear shapes 
    ax0=0.0
    ay0=0.0
    az0=0.0             ! ax0(0)=1.-ax => i   ax0(1)=ax => i+1 
    do n=1,nb
     r=p_acc(n,0)
     cn=xvp(n,3)/r
     r2c=xvp(n,1)*xvp(n,1)+xvp(n,2)*xvp(n,2)
     sn=sqrt(r2c)
     cnp=xvp(n,1)/sn
     snp=xvp(n,2)/sn
     sn=sn/r

     ax=xvp(n,0)
     ax=rx*ax-half
     i=int(ax)
     if(ax <=0.0)i=-1
     ax0(1)=ax-float(i)
     if(i==n1)ax0(1)=0.0
     ax0(0)=1.0-ax0(1)
!
     ay=ry*acos(cn)-half
     j=int(ay)
     if(ay <=0.0)j=-1
     ay0(1)=ay-float(j)
     ay0(0)=1.0-ay0(1)
     
     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi

     az=phi*rz
     k=int(az)
     az0(1)=az-float(k)
     az0(0)=1.0-az0(1)

     p_acc(n,1:3)=0.0
     do k1=0,1
      k2=k+k1
      if(k2 >=n3)k2=k2-n3 
      do j1=0,1
       j2=j+j1
       do i1=0,1
        i2=i+i1
        dvol=ax0(i1)*ay0(j1)*az0(k1)
        p_acc(n,1)=p_acc(n,1)+  &
                               Gr(i2,j2,k2)*dvol
        p_acc(n,2)=p_acc(n,2)+  &
                               Gth(i2,j2,k2)*dvol
        p_acc(n,3)=p_acc(n,3)+  &
                               Gph(i2,j2,k2)*dvol
       end do
      end do
     end do
    p_acc(n,2)=p_acc(n,2)/r
    p_acc(n,3)=p_acc(n,3)/(r*sn)

    fz=p_acc(n,1)*cn-p_acc(n,2)*sn
    fx=p_acc(n,1)*sn+p_acc(n,2)*cn
    p_acc(n,1)=fx*cnp-p_acc(n,3)*snp
    p_acc(n,2)=fx*snp+p_acc(n,3)*cnp
    p_acc(n,3)=fz
   end do
    
!---------------------
   case(2)               ! quadratic shapes ax(0:2)=> [i-1,i,i+1] 
    ax0=0.0
    ay0=0.0
!                      ax0(0)=> i-1  ax0(1)=> i  ax0(2)=> i+1
    do n=1,nb
     r=p_acc(n,0)
     cn=xvp(n,3)/r
     r2c=xvp(n,1)*xvp(n,1)+xvp(n,2)*xvp(n,2)
     sn=sqrt(r2c)
     cnp=xvp(n,1)/sn
     snp=xvp(n,2)/sn
     sn=sn/r

     ax=xvp(n,0)
     ax=rx*ax-half
     i=nint(ax)
     ax=ax-float(i)
     ax2=ax*ax
     ax0(1)=thr-ax2
     ax0(2)=0.5*(0.25+ax2+ax)
     if(i==n1)then
      ax0(2)=0.0
      ax0(1)=1.-abs(ax)
     endif
     ax0(0)=1.0-ax0(1)-ax0(2)
!
     ay=ry*acos(cn)-half
     j=nint(ay)
     ay=ay-float(j)
     ax2=ay*ay
     ay0(1)=thr-ax2
     ay0(2)=0.5*(0.25+ax2+ay)
     ay0(0)=1.0-ay0(1)-ay0(2)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=nint(az)
     az=az-float(k)
     ax2=az*az
     az0(1)=thr-ax2
     az0(2)=0.5*(0.25+ax2+az)
     az0(0)=1.0-az0(1)-az0(2)
     
     p_acc(n,1:3)=0.0
     do k1=0,2
      k2=k-1+k1
      if(k2 <0)k2=k2+n3 
      if(k2 > n3)k2=k2-n3 
      do j1=0,2
       j2=j-1+j1
       do i1=0,2
        i2=i-1+i1
        dvol=ax0(i1)*ay0(j1)*az0(k1)
        p_acc(n,1)=p_acc(n,1)+  &
                               Gr(i2,j2,k2)*dvol
        p_acc(n,2)=p_acc(n,2)+  &
                               Gth(i2,j2,k2)*dvol
        p_acc(n,3)=p_acc(n,3)+  &
                               Gph(i2,j2,k2)*dvol
       end do
      end do
     end do
!------------------ compose to cartesian coord.
    p_acc(n,2)=p_acc(n,2)/r
    p_acc(n,3)=p_acc(n,3)/(r*sn)

    fz=p_acc(n,1)*cn-p_acc(n,2)*sn
    fx=p_acc(n,1)*sn+p_acc(n,2)*cn
    p_acc(n,1)=fx*cnp-p_acc(n,3)*snp
    p_acc(n,2)=fx*snp+p_acc(n,3)*cnp
    p_acc(n,3)=fz
  end do

 end select

 end subroutine part_acc
!=============================
! Assigns grid-based potential to particles positions by interpolation
!============================================= 

 subroutine part_pot(pt,n1,n2,n3)
   integer,intent(in) :: n1,n2,n3

   integer :: n,i,i1,i2,j,j1,j2,k,k1,k2
    real,intent(in) :: pt(-1:n1+1,-1:n2,0:n3)
    real :: ax,ax2,ay,az,r2,r,phi
    real :: ax0(0:2),ay0(0:2),az0(0:2)
    real :: cn,sn,cnp,snp,fx,fz

!-------------------
    call inv_rad_map(nb)
!-------------------
   select case(ord)
!-----------------------
    case(1)               ! linear shapes 
    ax0=0.0
    ay0=0.0
    az0=0.0             ! ax0(0)=1.-ax => i   ax0(1)=ax => i+1 
    do n=1,nb
     r=p_acc(n,0)
     cn=xvp(n,3)/r

     ax=xvp(n,0)
     ax=rx*ax-half
     i=int(ax)
     if(ax<=0.0)i=-1
     ax0(1)=ax-float(i)
     ax0(0)=1.0-ax0(1)
!
     ay=ry*acos(cn)-half
     j=int(ay)
     if(ay <=0.0)j=-1
     ay0(1)=ay-float(j)
     ay0(0)=1.0-ay0(1)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=int(az)
     az0(1)=az-float(k)
     az0(0)=1.0-az0(1)

     p_acc(n,0)=0.0
     do k1=0,1
      k2=k+k1
      if(k2 >=n3)k2=k2-n3 
      do j1=0,1
       j2=j+j1
       do i1=0,1
        i2=i+i1
        p_acc(n,0)=p_acc(n,0)+   &
                               pt(i2,j2,k2)*ax0(i1)*ay0(j1)*az0(k1)
       end do
      end do
     end do
   end do
!---------------------
   case(2)               ! quadratic shapes 
    do n=1,nb
     r=p_acc(n,0)
     cn=xvp(n,3)/r
     ax=xvp(n,0)
     ax=rx*ax-half
     i=nint(ax)
     ax=ax-float(i)
     ax2=ax*ax
     ax0(1)=thr-ax2
     ax0(2)=0.5*(0.25+ax2+ax)
     if(i==n1)then
      ax0(2)=0.0
      ax0(1)=1.0-abs(ax)
     endif
     ax0(0)=1.0-ax0(1)-ax0(2)
!
     ay=ry*acos(cn)-half
     j=nint(ay)
     ay=ay-float(j)
     ax2=ay*ay
     ay0(1)=thr-ax2
     ay0(2)=0.5*(0.25+ax2+ay)
     ay0(0)=1.0-ay0(1)-ay0(2)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=nint(az)
     az=az-float(k)
     ax2=az*az
     az0(1)=thr-ax2
     az0(2)=0.5*(0.25+ax2+az)
     az0(0)=1.0-az0(1)-az0(2)
     
     p_acc(n,0)=0.0
     do k1=0,2
      k2=k-1+k1
      if(k2 <0)k2=k2+n3 
      if(k2 > n3)k2=k2-n3 
      do j1=0,2
       j2=j-1+j1
       do i1=0,2
        i2=i-1+i1
        p_acc(n,0)=p_acc(n,0)+  &
                               pt(i2,j2,k2)*ax0(i1)*ay0(j1)*az0(k1)
       end do
      end do
     end do
  end do

  end select
 end subroutine part_pot
!===================
! One dts time step for particles positions and velocity advance
!===============================
 subroutine pos_adv(dts)

  real,intent(in) :: dts
  integer :: ip

   do ip=1,nb
    xvp(ip,1:3)=xvp(ip,1:3)+dts*xvp(ip,4:6)
   end do
 end subroutine pos_adv

 subroutine vel_adv(dts)

  real,intent(in) :: dts
  integer :: ip

   do ip=1,nb
    xvp(ip,4:6)=xvp(ip,4:6)+dts*p_acc(ip,1:3)
   end do
 end subroutine vel_adv
!===================
! Assigns particles mass to grid points to set grid-based density field 
!===============================
 subroutine set_den(n1,n2,n3)
   integer,intent(in) :: n1,n2,n3

   integer :: n,i,i1,j,j1,j2,k,k1,k2
   real :: ax,ax2,ay,az,r2,r,phi,cn
   real :: ax0(0:2),ay0(0:2),az0(0:2)

    den=0.0
    call inv_rad_map(nb) 
!-------------------
   select case(ord)
!-----------------------
   case(1)               ! linear shapes 
    ax0=0.0
    ay0=0.0
    az0=0.0             ! ax0(0)=1.-ax => i   ax0(1)=ax => i+1 
    do n=1,nb
     r=p_acc(n,0)
     ax=xvp(n,0)
     ax=rx*ax-half
     i=int(ax)
     if(ax <=0.0)i=-1
     ax0(1)=ax-float(i)
     if(i==n1)ax0(1)=0.0
     ax0(0)=1.0-ax0(1)
!
     cn=xvp(n,3)/r
     ay=ry*acos(cn)-half
     j=int(ay)
     if(ay <= 0.0)j=-1
     ay0(1)=ay-float(j)
     ay0(0)=1.0-ay0(1)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=int(az)
     az0(1)=az-float(k)
     az0(0)=1.0-az0(1)

     do k1=0,1
      k2=k+k1
      if(k2 >=n3)k2=k2-n3 
      do j1=0,1
       j2=j+j1
       do i1=0,1
        den(i+i1,j2,k2)=den(i+i1,j2,k2)+ &
                               ax0(i1)*ay0(j1)*az0(k1)
       end do
      end do
     end do
    end do
!--------------------------------------------
   case(2)               ! quadratic splines: x=(x_p-x_i)/Delta x: [-1/2,1/2]
                         ! W(i)=3/4-x*x; 
                         !W(i+1)=1/2[3/2-|x-1|]^2; W(i-1)=1-W(i)-W(i+1)  
                         !---------------------------------------------------
    do n=1,nb
     r=p_acc(n,0)
     ax=xvp(n,0)
     ax=rx*ax-half
     i=nint(ax)
     ax=ax-float(i)
     ax2=ax*ax
     ax0(1)=thr-ax2
     ax0(2)=0.5*(0.25+ax2+ax)
     if(i==n1)then
      ax0(2)=0.0
      ax0(1)=1.0-abs(ax)
     endif
     ax0(0)=1.0-ax0(1)-ax0(2)
!
     cn=xvp(n,3)/r
     ay=ry*acos(cn)-half
     j=nint(ay)
     ay=ay-float(j)
     ax2=ay*ay
     ay0(1)=thr-ax2
     ay0(2)=0.5*(0.25+ax2+ay)
     ay0(0)=1.0-ay0(1)-ay0(2)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=nint(az)
     az=az-float(k)
     ax2=az*az
     az0(1)=thr-ax2
     az0(2)=0.5*(0.25+ax2+az)
     az0(0)=1.0-az0(1)-az0(2)

     do k1=0,2
      k2=k-1+k1
      if(k2 <0)k2=k2+n3 
      if(k2 >=n3)k2=k2-n3 
      do j1=0,2
       j2=j-1+j1
       do i1=0,2
        den(i-1+i1,j2,k2)=den(i-1+i1,j2,k2)+ &
                               ax0(i1)*ay0(j1)*az0(k1)
       end do
      end do
     end do
   end do
  end select
                                          ! redistribute on the effective grid
                                                 ! k=n3 
  call den_bd(n2/2,n3/2)           ! sums extra-boudary points  

  contains
   subroutine den_bd(nh2,nh3)
    integer,intent(in):: nh2,nh3
    integer :: j,j1,k,k1

    
    !den(-1:n1+1,-1:n2,0)=den(-1:n1+1,-1:n2,0)+den(-1:n1+1,-1:n2,n3)
                                                 ! i=n1+1
    den(n1,-1:n2,0:n3-1)=den(n1,-1:n2,0:n3-1)+den(n1+1,-1:n2,0:n3-1)

    do k=0,n3-1
     k1=k+nh3
     if(k1 >=n3)k1=k1-n3
     den(0:n1,0,k)=den(0:n1,0,k)+den(0:n1,-1,k1)
     den(0:n1,n2-1,k)=den(0:n1,n2-1,k)+den(0:n1,n2,k1)
     do j=0,n2-1
      j1=n2-1-j
      den(0,j,k)=den(0,j,k)+den(-1,j1,k1)
     end do
    end do
    den(0,0,0:n3-1)=den(0,0,0:n3-1)+den(-1,n2,0:n3-1)
    den(0,n2-1,0:n3-1)=den(0,n2-1,0:n3-1)+den(-1,-1,0:n3-1)
  end subroutine den_bd

 end subroutine set_den
!===================
! Assigns particles velocities to grid points to set grid-based velocity field 
!===============================
 subroutine set_part_vel(v1,v2,v3,n1,n2,n3)
   integer,intent(in) :: n1,n2,n3
   real,intent(inout) :: v1(-1:n1+1,-1:n2,0:n3)
   real,intent(inout) :: v2(-1:n1+1,-1:n2,0:n3)
   real,intent(inout) :: v3(-1:n1+1,-1:n2,0:n3)

   integer :: n,i,i1,j,j1,j2,k,k1,k2
   real :: ax,ax2,ay,az,r2,r,phi,cn
   real :: ax0(0:2),ay0(0:2),az0(0:2)

    v1=0.0
    v2=0.0
    v3=0.0
    call inv_rad_map(nb)
!-------------------
   select case(ord)
!-----------------------
   case(1)               ! linear shapes 
    ax0=0.0
    ay0=0.0
    az0=0.0             ! ax0(0)=1.-ax => i   ax0(1)=ax => i+1 
    do n=1,nb
     r=p_acc(n,0)
     ax=xvp(n,0)
     ax=rx*ax-half
     i=int(ax)
     ax=ax-float(i)
     ax0(1)=ax
     if(i==n1)ax0(1)=0.0
     if(ax <0.0)then
      i=-1
      ax0(1)=1.0          !unit weight at i=0 point
     endif
     ax0(0)=1.0-ax0(1)
!
     cn=xvp(n,3)/r
     ay=ry*acos(cn)-half
     j=int(ay)
     if(ay <=0.0)j=-1
     ay=ay-float(j)
     ay0(1)=ay
     ay0(0)=1.0-ay0(1)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=int(az)
     az=az-float(k)
     az0(1)=az
     az0(0)=1.0-az0(1)

     do k1=0,1
      k2=k+k1
      if(k2 >=n3)k2=k2-n3 
      do j1=0,1
       j2=j+j1
       do i1=0,1
        v1(i+i1,j2,k2)=v1(i+i1,j2,k2)+ &
                              xvp(n,4)*ax0(i1)*ay0(j1)*az0(k1)
        v2(i+i1,j2,k2)=v2(i+i1,j2,k2)+ &
                              xvp(n,5)*ax0(i1)*ay0(j1)*az0(k1)
        v3(i+i1,j2,k2)=v3(i+i1,j2,k2)+ &
                              xvp(n,6)*ax0(i1)*ay0(j1)*az0(k1)
       end do
      end do
     end do
    end do
!--------------------------------------------
   case(2)              
    do n=1,nb
     r=p_acc(n,0)
     ax=xvp(n,0)
     ax=rx*ax-half
     i=nint(ax)
     ax=ax-float(i)
     ax2=ax*ax
     ax0(1)=thr-ax2
     ax0(2)=0.5*(0.25+ax2+ax)
     if(i==n1)then
      ax0(2)=0.0
      ax0(1)=abs(ax)
     endif
     ax0(0)=1.0-ax0(1)-ax0(2)
!
     cn=xvp(n,3)/r
     ay=ry*acos(cn)-half
     j=nint(ay)
     ay=ay-float(j)
     ax2=ay*ay
     ay0(1)=thr-ax2
     ay0(2)=0.5*(0.25+ax2+ay)
     ay0(0)=1.0-ay0(1)-ay0(2)

     phi=atan2(xvp(n,2),xvp(n,1))
     if (phi<0.0) phi=2.0*pig+phi
     az=phi*rz
     k=nint(az)
     az=az-float(k)
     ax2=az*az
     az0(1)=thr-ax2
     az0(2)=0.5*(0.25+ax2+az)
     az0(0)=1.0-az0(1)-az0(2)

     do k1=0,2
      k2=k-1+k1
      if(k2 <0)k2=k2+n3 
      if(k2 >=n3)k2=k2-n3 
      do j1=0,2
       j2=j-1+j1
       do i1=0,2
        v1(i-1+i1,j2,k2)=v1(i-1+i1,j2,k2)+ &
                               xvp(n,4)*ax0(i1)*ay0(j1)*az0(k1)
        v2(i-1+i1,j2,k2)=v2(i-1+i1,j2,k2)+ &
                               xvp(n,5)*ax0(i1)*ay0(j1)*az0(k1)
        v3(i-1+i1,j2,k2)=v3(i-1+i1,j2,k2)+ &
                               xvp(n,6)*ax0(i1)*ay0(j1)*az0(k1)
       end do
      end do
     end do
   end do
  end select
                                          ! redistribute on the effective grid
                                                 ! k=n3 
  call vel_bd(n2/2,n3/2)           ! sum extra-boudary points  

  contains
   subroutine vel_bd(nh2,nh3)
    integer,intent(in):: nh2,nh3
    integer :: j,j1,k,k1

    
    v1(-1:n1+1,-1:n2,0)=v1(-1:n1+1,-1:n2,0)+v1(-1:n1+1,-1:n2,n3)
    v2(-1:n1+1,-1:n2,0)=v2(-1:n1+1,-1:n2,0)+v2(-1:n1+1,-1:n2,n3)
    v3(-1:n1+1,-1:n2,0)=v3(-1:n1+1,-1:n2,0)+v3(-1:n1+1,-1:n2,n3)
                                                 ! i=n1+1
    v1(n1,-1:n2,0:n3-1)=v1(n1,-1:n2,0:n3-1)+v1(n1+1,-1:n2,0:n3-1)
    v2(n1,-1:n2,0:n3-1)=v2(n1,-1:n2,0:n3-1)+v2(n1+1,-1:n2,0:n3-1)
    v3(n1,-1:n2,0:n3-1)=v3(n1,-1:n2,0:n3-1)+v3(n1+1,-1:n2,0:n3-1)

    do k=0,n3-1
     k1=k+nh3
     if(k1 >=n3)k1=k1-n3
     v1(0:n1,0,k)=v1(0:n1,0,k)+v1(0:n1,-1,k1)
     v1(0:n1,n2-1,k)=v1(0:n1,n2-1,k)+v1(0:n1,n2,k1)
     v2(0:n1,0,k)=v2(0:n1,0,k)+v2(0:n1,-1,k1)
     v2(0:n1,n2-1,k)=v2(0:n1,n2-1,k)+v2(0:n1,n2,k1)
     v3(0:n1,0,k)=v3(0:n1,0,k)+v3(0:n1,-1,k1)
     v3(0:n1,n2-1,k)=v3(0:n1,n2-1,k)+v3(0:n1,n2,k1)
     do j=0,n2-1
      j1=n2-1-j
      v1(0,j,k)=v1(0,j,k)+v1(-1,j1,k1)
      v2(0,j,k)=v2(0,j,k)+v2(-1,j1,k1)
      v3(0,j,k)=v3(0,j,k)+v3(-1,j1,k1)
     end do
    end do
    v1(0,0,0:n3-1)=v1(0,0,0:n3-1)+v1(-1,n2,0:n3-1)
    v1(0,n2-1,0:n3-1)=v1(0,n2-1,0:n3-1)+v1(-1,-1,0:n3-1)
  end subroutine vel_bd

 end subroutine set_part_vel

 end module particles
!----------------------------------
!last change 5 sept 2008
! in subroutines part_acc part_pot _set_den set_part_vel
! corrected assignment of phi coordinate
!     phi=atan2(xvp(n,2),xvp(n,1))
!     if (phi<0.0) phi=2.0*pig+phi
